/*
********************************************************************************
cmosio - CMOS I/O

Read or write CMOS NVRAM of the x86-type PC architecture.
Establish requirements for access control to the I/O port memory addresses,
and provide an option for each of the access methods.

More information:
http://www.faqs.org/docs/Linux-mini/IO-Port-Programming.html
http://www.codepedia.com/1/CMOS_C

Another CMOS memory map:
http://www.bioscentral.com/misc/cmosmap.htm

Compile Linux (runs on 32bit + 64bit):
gcc -Wall -m32 -march=i386 -mtune=i386 -g -o cmosio cmosio.c

Compile FreeBSD (which also needs GNU getopt()):
gcc -Wall -I getopt -g -c -o getopt.o getopt/getopt.c
gcc -Wall -g getopt.o -o cmosio cmosio.c
gcc -Wall -g getopt.o -o cmosio-static -static cmosio.c

Notes:

  * This program must be run as root when not operating only on disk files.

Linux:

  * There are 3 different ways to access I/O addresses, but 2 of them use the
    same principle - assembler instructions, with 1 using the declarations from
    a system header file.

  * When accessing IO memory through /dev/port, the file must be opened
    read/write even when only reading the CMOS NVRAM, because the NVRAM address
    to read first needs to be written to another I/O-address register.

  * The default I/O (direct) no longer works on Linux (openSUSE 11.4).
    The asm I/O works (it doesn't use the blocking I/O).
    Both fail the cli unless ioperm() is called appropriately.

  * Instructions cli, sti segfault if current privilege level is less than
    IO privilege level (even when run as root).
    http://classic-web.archive.org/web/20070802184708/http://www.x86.org/articles/pvi1/pvi1.htm

    It's probably a very bad idea to disable interrupts.

FreeBSD:

  * Use inb, outb assembler instructions. To be granted access, either hold the
    device special file /dev/io open for reading ("r"), or call
    i386_set_ioperm().
    http://www.cz.freebsd.org/pub/FreeBSD-cvs/gnats/threads/116181
    http://w140.com/kurt/freebsd_byte_io.html

Copyright (C) by Volker Kuhlmann, http://volker.top.geek.nz/contact.html
All rights reserved.
Released under the terms of the GNU General Public License (GPL) Version 2.
See http://www.gnu.org/ for details.

Volker Kuhlmann
  30 Dec 2010  Proof of concept.
  01 Jan 2011  Make a usable application program.
  29 May 2011  FreeBSD proof of concept.
  11 Jun 2011  Finish for Linux. Start on port to FreeBSD.
  23 Jul 2011  Finish FreeBSD version.
********************************************************************************
*/
#define VERSION "1.3  23 Jul 2011"


// Define this for debugging output, and to suppress potentially dangerous
// read/write operations to the CMOS area.
//#define USE_DEBUG

// Define this to compile for FreeBSD.
//#define FREEBSD
#ifdef __FreeBSD__  // Define automatically.
#define FREEBSD
#endif


/* Access methods to use for port I/O. */

// Device special file for accessing CMOS NVRAM. Works on Linux.
// FreeBSD: Device special file to get access permission.
#define USE_DEV_SPECIAL

// Assembler instructions for port I/O. Should always work.
#define USE_ASM_IO

// Use outb(), inb() from Linux sys/io.h. N/A FreeBSD.
#ifndef FREEBSD
#define USE_OUTB_INB
#endif


/* Other config values. */

// Device special file for accessing CMOS NVRAM.
#ifdef USE_DEV_SPECIAL
  #ifdef FREEBSD
    // Open this for reading to get permission for inb, outb.
    #define DEV_SPECIAL "/dev/io"
  #else
    // The content of this file is mapped to the I/O memory.
    // Use fseek(); similar to /proc/kcore.
    #define DEV_SPECIAL "/dev/port"
  #endif
#endif

// IO privilege level to use. Must be 3 for cli(), sti() to not segfault.
// 0 = don't disable interrupts.
#define IOPL 0

// CMOS NVRAM size. This MUST be no larger than 128!
#define CMOSSIZE 128

// Number of bytes at CMOS start address used by clock and alarm times.
#define CMOSTIMES 10

// Debugging output. Define DEBUG() to nothing to suppress.
#ifdef USE_DEBUG
#define DEBUG(toks) toks // suppress
#else
#define DEBUG(toks)
#endif

// Help text.
const char const * helptext = {
"Usage: cmosio [OPTIONS] ACTION\n"
"Version: " VERSION "\n"
"Copyright (C) 2010-2011 Volker Kuhlmann\n"
"\n"
"Actions:\n"
" -r FILE  Read CMOS, write to disk file FILE.\n"
" -w FILE  Read disk file FILE, write to CMOS (except clock and alarm times).\n"
" -c FILE1 FILE2  Compare the two CMOS content files.\n"
" -d       Dump contents of CMOS in readable text.\n"
" -d FILE  As -d, but dump FILE instead.\n"
" -0 ADDR MASK  Clear all bits in CMOS location ADDR that are set in MASK.\n"
"                 ADDR must be between 0 and 127. MASK is 8 bit.\n"
" -1 ADDR MASK  Set all bits in CMOS location ADDR that are set in MASK.\n"
" -s ADDR VALUE  Write VALUE into CMOS location ADDR.\n"
" -w FILE -0 ADDR MASK  As above, operate on FILE instead of CMOS.\n"
" -w FILE -1 ADDR MASK  As above, operate on FILE instead of CMOS.\n"
" -h  Show help\n"
"\n"
"Options:\n"
#ifdef FREEBSD
  // Access is always through assembler instructions.
  #ifdef USE_DEV_SPECIAL
  " -D  Access permissions via device special file '" DEV_SPECIAL "'.\n"
  #endif
  #ifdef USE_ASM_IO
  " -A  Access permissions via i386_set_ioperm().\n"
  #endif
#else
  #ifdef USE_DEV_SPECIAL
  " -D  Access CMOS via device special file '" DEV_SPECIAL "'.\n"
  #endif
  #ifdef USE_ASM_IO
  " -A  Access CMOS via assembler instructions.\n"
  #endif
#endif
#ifdef USE_OUTB_INB
" -O  Access CMOS via outb() function.\n"
#endif
" -t   With -w: write clock time as well\n"
"\n"
"Accepted formats for ADDR, MASK are decimal, binary (leading \"0b\"),\n"
"octal (leading \"0o\"), hexadecimal (leading \"0x\").\n"
"\n"
"Read or write operations to CMOS NVRAM require root privileges!\n"
"Insufficient privileges end with a segmentation fault or bus error.\n"
};


/* Header files */

// Standard C
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#ifdef USE_ASM_IO
#include <stdint.h>
#endif

// Unix
#include <unistd.h>
#include <getopt.h>

#ifdef USE_DEV_SPECIAL
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
  #ifdef FREEBSD
    #include <machine/sysarch.h>   // search in /usr/include/ for IOPL
  #else
    #include <sys/io.h>	// ioperm(), iopl()
  #endif
#endif

#ifdef USE_OUTB_INB
#include <sys/io.h>
#endif


/* Types & Variables */

// Exit codes of function return values.
enum exitcode_t {
    EXIT_OK,
    EXIT_OPENCLOSE,
    EXIT_IO,
    EXIT_CMDARG,
    EXIT_BUG = 9,
};

// CMOS access method.
enum method_t {
    METHOD_NONE,
    METHOD_DEV,     	    	    	// device special file
    METHOD_ASM,     	    	    	// assembler
    METHOD_OUTB,    	    	    	// outb(), inb()
};

// Buffer for CMOS NVRAM data.
unsigned char cmos[CMOSSIZE];

// Command line parameters.
struct conf_t {
    int writeclock;
    char cmd;
    enum method_t method;
    char * file1;
    char * file2;
    int addr;
    int mask;
} conf = {
    0,
    '\0',
    METHOD_NONE,
    NULL,
    NULL,
    -1,
    0,
};


/*******************************************************************************
Helper Functions
*******************************************************************************/

// Convert a byte into a binary string. Buffer must be 9 bytes!
// No, printf() doesn't have it.
char * itob(char val, char * buffer)
{
    int i;
    char * start = buffer;
    
    for (i = 0; i < 8; i++) {
	*buffer++ = (val & 0x80) ? '1' : '0';
	val <<= 1;
    }
    *buffer = '\0'; // terminate 8-character string
    
    return start;
}

// Convert a string into an int between 0 and 255.
// Return -1, or the numerical value of the string.
int strtoint(char * str)
{
    char * s;
    long l;
    int base;
    
    errno = 0;
    if (strncmp(str, "0b", 2) == 0)
    	base = 2;
    else if (strncmp(str, "0x", 2) == 0)
    	base = 16;
    else if (strncmp(str, "0o", 2) == 0)
    	base = 8;
    else
    	base = 10;
    if (base != 10)
    	str += 2; // skip prefix characters
    l = strtol(str, &s, base);
    DEBUG( printf("strtoint() '%s', '%s', %i\n", str, s, errno); )
    if (errno != 0 || *s != '\0' || s == str)
    	return -1; // conversion error in strtol()
    else if (l < 0 && l > 255)
	return -1; // value out of our range
    else
    	return l; // conversion and value OK
}

// Show error text and return the supplied exit code.
int err(int exitcode, const char * str)
{
    fprintf(stderr, "%s\n", str);
    return exitcode;
}

// Show missing method error and return error code.
int err_method(void)
{
    return err(EXIT_CMDARG, "Error: No CMOS access method given!");
}


/*******************************************************************************
CMOS NVRAM read/write
*******************************************************************************/

#ifdef USE_DEV_SPECIAL

/*
 Read/write CMOS NVRAM through a device special file, usually /dev/port.
 This operation should always work, but only on Linux.
 On other systems it will compile, but produce an error opening the file.
 
 Note FreeBSD: File needs to be open to get access (as root) to I/O memory
 addresses, but reading/writing has no purpose.
*/

#ifdef FREEBSD
int saved_devspecial_fd;
#endif

// Open device special file, print any error.
// Return fd or -1.
int open_devspecial()
{
    int fd;
    
    fd = open(DEV_SPECIAL, O_RDWR);
    if (fd == -1) {
	perror("Error open(\"" DEV_SPECIAL "\")");
    }
    DEBUG( return 0; )
    return fd;
}

// Close device special file, print any error.
// Return 0 or -1.
int close_devspecial(int fd)
{
    int e;
    
    e = close(fd);
    if (e) {
	perror("Error close()");
    	DEBUG( return 0; )
	return -1;
    } else
    	return 0;
}

// Read a byte from CMOS NVRAM using a device special file.
// Return exit code.
// Very inefficient, but it's <= 128 bytes to read.
int getCMOS_devspecial(int reg, int * value)
{
    char c;
    int fd;

    fd = open_devspecial();
    DEBUG( *value = 0x3D; return EXIT_OK; )
    if (fd == -1)
	return EXIT_OPENCLOSE;

    lseek(fd, 0x70, SEEK_SET);
    if (errno) {
	perror("Error lseek()");
	return EXIT_IO;
    }
    c = reg;
    write(fd, &c, 1);
    if (errno) {
	perror("Error write()");
	return EXIT_IO;
    }
    read(fd, &c, 1);  // 0x71
    if (errno) {
	perror("Error read()");
	return EXIT_IO;
    }
    *value = c;

    if (close_devspecial(fd) == -1)
	return EXIT_OPENCLOSE;
    
    return EXIT_OK;
}

// Write a byte to CMOS NVRAM using a device special file.
// Return exit code.
// Very inefficient, but it's <= 128 bytes to write.
int setCMOS_devspecial(int reg, int value)
{
    char c;
    int fd;

    DEBUG( printf("set 0x%02x -> 0x%02X\n", reg, value); )
    fd = open_devspecial();
    DEBUG( return EXIT_OK; )
    if (fd == -1)
	return EXIT_OPENCLOSE;

    lseek(fd, 0x70, SEEK_SET);
    if (errno) {
	perror("Error lseek()");
	return EXIT_IO;
    }
    c = reg;
    write(fd, &c, 1);
    if (errno) {
	perror("Error write()");
	return EXIT_IO;
    }
    c = value;
    write(fd, &c, 1);  // 0x71
    if (errno) {
	perror("Error write()");
	return EXIT_IO;
    }

    if (close_devspecial(fd) == -1)
	return EXIT_OPENCLOSE;
    
    return EXIT_OK;
}

#endif // USE_DEV_SPECIAL


#if defined(USE_ASM_IO) || defined(USE_OUTB_INB)

// Disable interrupts. (Warning: must be for as short as possible!)
inline void cli() {
    // Note cli segfaults unless run as root.
    // cli also segfaults if run as root, with IOPL < CPL (IO privilege level,
    // current privilege level).
    // Needs iopl(3) to not segfault.
    
    // This is probably also dangerous, so don't use it.
    //asm("cli");
}
// Enable interrupts.
inline void sti() {
    // Same as for cli above applies here.
    
    //asm("sti");
}

#endif


#ifdef USE_ASM_IO

// Write I/O-port byte.
// (This is very similar to outb(), which does
//   ("outb %b0,%w1": :"a" (__value), "Nd" (__port))    in sys/io.h)
static __inline__ 
void outportb(uint8_t value, uint16_t port)
{
  __asm__ volatile ("outb %0,%1"
		    ::"a" ((char) value), "d"((uint16_t) port));
}
// Read I/O-port byte.
// (This is very similar to outb(), which does
//   ("inb %w1,%0":"=a" (_v):"Nd" (__port))    in sys/io.h)
static __inline__ 
uint8_t inportb(uint16_t port)
{
  uint8_t _v;
  __asm__ volatile ("inb %1,%0"
			 :"=a" (_v):"d"((uint16_t) port));
  return _v;
}
// Read CMOS NVRAM byte.
inline int getCMOS_asm(int reg)
{
    int i;
    
    DEBUG( return 0x5D; )
    cli();
    outportb(reg, 0x70);
    i = inportb(0x71);
    sti();
    return i;
}
// Write CMOS NVRAM byte.
inline void setCMOS_asm(int reg, int value)
{
    DEBUG( printf("set 0x%02x -> 0x%02X\n", reg, value); return; )
    cli();
    outportb(reg, 0x70);
    outportb(value, 0x71);
    sti();
}

#endif // USE_ASM_IO


#ifdef USE_OUTB_INB // Use outb_p(), inb_p()

// outb_p(), inb_p() wait for the operation to finish.
// Note that they also access IO port 0x80!

inline int getCMOS_outb(int reg)
{
    int i;
    
    DEBUG( return 0x5D; )
    cli();
    outb_p(reg, 0x70);
    i = inb_p(0x71);
    sti();
    return i;
}
inline void setCMOS_outb(int reg, int value)
{
    DEBUG( printf("set 0x%02x -> 0x%02X\n", reg, value); return; )
    cli();
    outb_p(reg, 0x70);
    outb_p(value, 0x71);
    sti();
}

#endif // Use outb_p(), inb_p()


// Get privileges.
// Return error code.
int getprivs(void)
{
    int ret = EXIT_OK;

#ifdef FREEBSD
  #if defined(USE_ASM_IO) || defined(USE_DEV_SPECIAL)
    if (conf.method == METHOD_DEV) {
    	saved_devspecial_fd = open_devspecial();
    } else {
	ret = i386_set_ioperm(0x70, 2, 3);
	if (ret != EXIT_OK) {
    	    perror("Failed to set I/O port privileges");
    	    return ret;
	}
    }
  #endif
#else
  #if defined(USE_ASM_IO) || defined(USE_OUTB_INB)
    if (conf.method == METHOD_ASM || conf.method == METHOD_OUTB) {
	ret = ioperm(0x70, 2, 3); // 0x70 CMOS register, 0x71 value
    	#ifdef USE_OUTB_INB
	ret = ioperm(0x80, 1, 3); // outb_p(), inb_p()
    	#endif
	if (ret != EXIT_OK) {
    	    perror("Failed to set I/O port privileges");
    	    return ret;
	}
	// For disabling interrupts:
	ret = iopl(IOPL);
	if (ret != EXIT_OK) {
    	    perror("Failed to set I/O privilege level");
    	    return ret;
	}
    }
  #endif
#endif
    
    return ret;
}

// Drop privileges.
// Return error code.
int dropprivs(void)
{
    int ret = EXIT_OK;

#ifdef FREEBSD
  #if defined(USE_ASM_IO) || defined(USE_DEV_SPECIAL)
    if (conf.method == METHOD_DEV) {
    	close_devspecial(saved_devspecial_fd);
    } else {
    	ret = i386_set_ioperm(0x70, 2, 0);
	if (ret != EXIT_OK) {
    	    perror("Failed to set I/O port privileges");
	}
    }
  #endif
#else
  #if defined(USE_ASM_IO) || defined(USE_OUTB_INB)
    // Release privileges.
    if (conf.method == METHOD_ASM || conf.method == METHOD_OUTB) {
    	ioperm(0x70, 2, 0);
    	#ifdef USE_OUTB_INB
	ioperm(0x80, 1, 0);
    	#endif
	iopl(0);
    }
  #endif
#endif

    return ret;
}

// Read CMOS NVRAM byte by selected method.
// Return error code.
int getCMOS(int reg, int * value)
{
    int ret = EXIT_OK;
    
    switch (conf.method) {
#ifdef USE_DEV_SPECIAL
    case METHOD_DEV:
    	#ifdef FREEBSD
    	*value = getCMOS_asm(reg);
	#else
    	ret = getCMOS_devspecial(reg, value);
	#endif
	break;
#endif
#ifdef USE_ASM_IO
    case METHOD_ASM:
    	*value = getCMOS_asm(reg);
    	break;
#endif
#ifdef USE_OUTB_INB
    case METHOD_OUTB:
    	*value = getCMOS_outb(reg);
    	break;
#endif
    default:
    	ret = err_method();
    }
    
    return ret;
}

// Write CMOS NVRAM byte by selected method.
// Return error code.
int setCMOS(int reg, int value)
{
    int ret = EXIT_OK;
    
    switch (conf.method) {
#ifdef USE_DEV_SPECIAL
    case METHOD_DEV:
    	#ifdef FREEBSD
    	setCMOS_asm(reg, value);
	#else
    	ret = setCMOS_devspecial(reg, value);
	#endif
	break;
#endif
#ifdef USE_ASM_IO
    case METHOD_ASM:
    	setCMOS_asm(reg, value);
    	break;
#endif
#ifdef USE_OUTB_INB
    case METHOD_OUTB:
    	setCMOS_outb(reg, value);
    	break;
#endif
    default:
    	ret = err_method();
    }
    
    return ret;
}

// Read CMOS NVRAM into buffer.
// Return exit code.
int readCMOS(void)
{
    int i;
    int ret;

    // Get privileges.
    ret = getprivs();
    if (ret != EXIT_OK)
    	return ret;

    for (i = 0; i < CMOSSIZE; i++) {
	// Very inefficient, but it's <= 128 bytes.
	ret = getCMOS(i, (int *) &cmos[i]);
	if (ret != EXIT_OK)
    	    return ret;
    }

    // Release privileges.
    ret = dropprivs();

    return ret;
}

// Write buffer to CMOS NVRAM.
// Return exit code.
int writeCMOS(void)
{
    int i;
    int ret;

    // Get privileges.
    ret = getprivs();
    if (ret != EXIT_OK)
    	return ret;

    for (i = 0; i < CMOSSIZE; i++) {
    	// Skip the bytes for clock and alarm times unless requested.
    	if (i < CMOSTIMES && !conf.writeclock)
	    continue;

	// very inefficient, but it's <= 128 bytes
	ret = setCMOS(i, cmos[i]);
	if (ret != EXIT_OK) {
	    (void) dropprivs();
    	    return ret;
	}
    }
    // Release privileges.
    ret = dropprivs();

    return ret;
}


/*******************************************************************************
Disk file I/O
*******************************************************************************/

// Read a file into buffer.
// Return exit code.
int readBufferFromFile(const char * filename, unsigned char * buffer)
{
    size_t len;
    FILE * file;
    
    file = fopen(filename, "rb");
    if (!file) {
    	perror("Error opening file");
	return EXIT_OPENCLOSE;
    }
    len = fread(buffer, 1, CMOSSIZE, file);
    if (len < CMOSSIZE) {
	fprintf(stderr, "Read fewer than %d bytes from file '%s'!\n",
		CMOSSIZE, filename);
	return EXIT_IO;
    }
    fclose(file);
    return EXIT_OK;
}

// Write buffer to a file.
// Return exit code.
int writeBufferToFile(const char * filename, const unsigned char * buffer)
{
    size_t len;
    FILE * file;
    
    file = fopen(filename, "wb");
    if (!file) {
    	perror("Error opening file");
	return EXIT_OPENCLOSE;
    }
    len = fwrite(buffer, 1, CMOSSIZE, file);
    if (len < CMOSSIZE) {
	fprintf(stderr, "Wrote fewer than %d bytes to file '%s'!\n",
		CMOSSIZE, filename);
	return EXIT_IO;
    }
    fclose(file);
    return EXIT_OK;
}


/*******************************************************************************
Actions
*******************************************************************************/

// Display help text.
void showhelp(FILE * output)
{
    fprintf(output, helptext);
}

// Compare two CMOS NVRAM files and output the differences.
// Return exit code.
int compareCMOSfiles()
{
    unsigned char buffer1[CMOSSIZE], buffer2[CMOSSIZE];
    char bin1[9], bin2[9];
    int ret, i;
    int headerprinted = 0;
    
    ret = readBufferFromFile(conf.file1, buffer1);
    ret = readBufferFromFile(conf.file2, buffer2);
    for (i = 0; i < CMOSSIZE; i++) {
	if (buffer1[i] != buffer2[i]) {
	    if (!headerprinted) {
	    	printf("Addr  file1 file2  file1      file2\n");
		headerprinted = 1;
	    }
	    printf("0x%02x  0x%02X   0x%02X  0b%s 0b%s\n",
		    i, 
		    buffer1[i], buffer2[i], 
		    itob(buffer1[i], bin1), itob(buffer2[i], bin2));
	}
    }
    return ret;
}

// Modify a byte in CMOS NVRAM with a bitwise logic operation.
// Return exit code.
int modifyCMOS(void)
{
    unsigned char byte, newbyte;
    char b1[9], b2[9], b3[9], buf[80];
    FILE * file;
    
    // range check
    if (conf.addr < 0 || conf.addr >= CMOSSIZE) {
    	sprintf(buf, "CMOS address must be 0-%d!", CMOSSIZE-1);
    	return err(EXIT_CMDARG, buf);
    }
    if (conf.mask < 0 || conf.mask > 255)
    	return err(EXIT_CMDARG, "Bit operation mask must be 0-255!");
    
    // Read byte from CMOS or file.
    if (conf.file1 != NULL) {
    	int c;
	file = fopen(conf.file1, "r+b");
	if (!file) {
    	    perror("Error opening file");
	    return EXIT_OPENCLOSE;
	}
    	if (fseek(file, conf.addr, SEEK_SET) == -1) {
	    perror("Error seeking in file");
	    return EXIT_IO;
	}
	c = fgetc(file);
	if (c == EOF)
	    return err(EXIT_IO, "Error reading byte");
	byte = c;
    } else {
	// Get privileges.
	int ret = getprivs();
	if (ret != EXIT_OK)
    	    return ret;
	if (getCMOS(conf.addr, (int *) &byte))
    	    return EXIT_IO;
    }

    // Modify byte.
    switch(conf.cmd) {
    case '0':
    	newbyte = byte & ~conf.mask;
    	printf("CMOS 0x%02x,%-3d"
	       "  0x%02X,0b%s,%-3d AND %s -> 0x%02X,0b%s,%-3d\n",
	    conf.addr, conf.addr, 
	    byte, itob(byte, b1), byte,
	    itob(~conf.mask, b2), 
	    newbyte, itob(newbyte, b3), newbyte);
    	break;
    case '1':
    	newbyte = byte | conf.mask;
    	printf("CMOS 0x%02x,%-3d"
	       "  0x%02X,0b%s,%-3d OR  %s -> 0x%02X,0b%s,%-3d\n",
	    conf.addr, conf.addr, 
	    byte, itob(byte, b1), byte,
	    itob(conf.mask, b2), 
	    newbyte, itob(newbyte, b3), newbyte);
    	break;
    default:
    	fprintf(stderr, "Ooops - can't do that!\n");
	return EXIT_BUG;
    }

    // Write byte to CMOS or file.
    if (conf.file1 != NULL) {
    	if (fseek(file, conf.addr, SEEK_SET) == -1) {
	    perror("Error seeking in file");
	    return EXIT_IO;
	}
	if (fputc(newbyte, file) != newbyte) {
	    perror("Error writing byte to file");
	    return EXIT_IO;
	}
	fclose(file);
    } else {
	if (setCMOS(conf.addr, newbyte)) {
	    (void) dropprivs();
    	    return EXIT_IO;
	}
    	return dropprivs();
    }
    
    return EXIT_OK;
}

// Write a single byte to a CMOS NVRAM location.
int writeCMOSbyte(void)
{
    char b1[9], buf[80];
    int ret;

    // range check
    if (conf.addr < 0 || conf.addr >= CMOSSIZE) {
    	sprintf(buf, "CMOS address must be 0-%d!", CMOSSIZE-1);
    	return err(EXIT_CMDARG, buf);
    }
    if (conf.mask < 0 || conf.mask > 255)
    	return err(EXIT_CMDARG, "Byte value must be 0-255!");
    
    printf("CMOS 0x%02x,%-3d"
	   "  -> 0x%02X,0b%s,%-3d\n",
	conf.addr, conf.addr, 
	conf.mask, itob(conf.mask, b1), conf.mask);

    ret = getprivs();
    if (ret != EXIT_OK)
    	return ret;

    if (setCMOS(conf.addr, conf.mask)) {
    	(void) dropprivs();
    	return EXIT_IO;
    }
    ret = dropprivs();
    
    return ret;
}

// Dump CMOS content, from CMOS NVRAM or file.
int dumpCMOS(void)
{
    char b1[9];
    int ret, i;

    if (conf.file1)
    	ret = readBufferFromFile(conf.file1, cmos);
    else
    	ret = readCMOS();
    if (ret != EXIT_OK)
    	return ret;

    for (i = 0; i < CMOSSIZE; i++) {
	printf("CMOS 0x%02x,%-3d"
	       "  0x%02X,0b%s,%-3d\n",
	    i, i, 
	    cmos[i], itob(cmos[i], b1), cmos[i]);
    }
    return ret;
}


/*******************************************************************************
Command line parsing
*******************************************************************************/

// Parse command line arguments.
// Return exit code.
int parse(int argc, char *argv[])
{
    int opt;
    
    while ((opt = getopt(argc, argv, "htDAOr:w:c:0:1:s:d::")) != -1) {
	switch (opt) {
	case 't':
            conf.writeclock = 1;
            break;
	case 'r':
	case 'w':
	case 'c':
            conf.cmd = opt;
	    conf.file1 = optarg;
            break;
	case '0':
	case '1':
	case 's':
            conf.cmd = opt;
	    conf.addr = strtoint(optarg);
	    if (conf.addr == -1)
	    	return EXIT_CMDARG;
            break;
	case 'd':
	case 'h':
	    conf.cmd = opt;
	    if (optarg != NULL)
	    	conf.file1 = optarg;
	    DEBUG( printf("opt=%c, optarg=%s\n", opt, optarg); )
	    break;
#ifdef USE_DEV_SPECIAL
	case 'D':
	    conf.method = METHOD_DEV;
	    break;
#endif
#ifdef USE_ASM_IO
	case 'A':
	    conf.method = METHOD_ASM;
	    break;
#endif
#ifdef USE_OUTB_INB
	case 'O':
	    conf.method = METHOD_OUTB;
	    break;
#endif
	default: /* '?' */
	    showhelp(stderr);
            return(EXIT_CMDARG);
	}
    }

    // Look for and store second argument for these actions.
    // (strchr() finds '\0' in string termination - d'oh!)
    if (strchr("c01s", conf.cmd) != NULL && conf.cmd != '\0') {
	if (argc > optind) {
	    switch (conf.cmd) {
	    case 'c':
	    	conf.file2 = argv[optind];
    		break;
	    case '0':
	    case '1':
	    case 's':
		conf.mask = strtoint(argv[optind]);
		if (conf.mask == -1)
	    	    return EXIT_CMDARG;
    		break;
	    default:
	    	;
	    }
	} else {
	    fprintf(stderr,
	    	"Expected second argument for at least one option!\n"
		"Call with -h for help.\n");
	    return(EXIT_CMDARG);
	}
    }

    // Look for and store optional arguments for these actions.
    // getopt() sucks, "d::" only finds the argument with "-darg", not "-d arg".
    if (strchr("d", conf.cmd) != NULL && conf.cmd != '\0') {
	if (argc > optind) {
	    switch (conf.cmd) {
	    case 'd':
	    	if (conf.file1 == NULL)
		    conf.file1 = argv[optind];
    		break;
	    default:
	    	;
	    }
	}
    }

    return(EXIT_OK);
}


/*******************************************************************************
Main
*******************************************************************************/

int main(int argc, char **argv)
{
    int ret;
    
    ret = parse(argc, argv);
    DEBUG( printf("clock=%d; cmd=%c; file1=%s; file2=%s; addr=%d; mask=%d\n", 
	    conf.writeclock, conf.cmd, conf.file1, conf.file2,
	    conf.addr, conf.mask);
    )
    if (ret != EXIT_OK)
	return(ret);
    
    ret = EXIT_BUG;
    switch (conf.cmd) {
    case 'r':
	ret = readCMOS();
	if (ret == EXIT_OK)
	    ret = writeBufferToFile(conf.file1, cmos);
	break;
    case 'w':
	ret = readBufferFromFile(conf.file1, cmos);
	if (ret == EXIT_OK)
	    ret = writeCMOS();
	break;
    case 'c':
	ret = compareCMOSfiles();
	break;
    case '0':
    case '1':
	ret = modifyCMOS();
	break;
    case 's':
	ret = writeCMOSbyte();
	break;
    case 'd':
    	ret = dumpCMOS();
	break;
    case 'h':
	showhelp(stdout);
	ret = EXIT_OK;
	break;
    case '\0':
    	ret = EXIT_CMDARG; // no action found
	// fall through
    default:
	showhelp(stderr);
    }

    return ret;
}

/******************************************************************************/
